__doc__ = """CachePool module"""
__version__ = '0.2.4a'

from Globals import HTMLFile, HTML, MessageDialog, Persistent
#from OFS.ZDOM import ElementWithTitle
#from OFS.SimpleItem import Item_w__name__
from AccessControl.Role import RoleManager
try:  #Zope 2.1.6
  from AccessControl.User import verify_watermark
except:
  #Zope 2.2.0
  #Still working on this
  pass
import OFS.SimpleItem
import Acquisition
import urllib
import time
from string import split, join
from thread import allocate_lock

manage_addCachePoolForm = HTMLFile('CachePoolAdd', globals())

def manage_addCachePool(self,id,title,DocToParse,TTL=60,CacheSlots=5,REQUEST=None):
  """Add CachePool Instance"""
  self._setObject(id,CachePool(id,title,DocToParse,TTL,CacheSlots))
  if REQUEST is not None:
    return self.manage_main(self, REQUEST)

class CachePool(
  HTML,
  Persistent,
  Acquisition.Implicit,
  RoleManager,
  OFS.SimpleItem.Item,
  ):
  "Cache Pool product"

  meta_type = 'Cache Pool'
  _proxy_roles=()
  index_html=None # Prevent accidental acquisition

  meta_types = ({'permission': 'Add CachePool', 'name': 'CachePool', 'product':
'CachePool', 'action': 'manage_addProduct/CachePool/CachePoolAdd'})

  try:
    #Zope 2.2
    manage_options=(
      (
        {'label':'Properties','action':'manage_editForm'},
        {'label':'View Cache','action':'manage_viewCache'},
        )
      +RoleManager.manage_options
      +Item_w__name__.manage_options
      )
  except:
    #Zope 2.1.6
    manage_options=(
      {'label':'Properties','action':'manage_editForm'},
      {'label':'View Cache','action':'manage_viewCache'},
      {'label':'Security','action':'manage_access'},
    )

  __ac_permissions__=(
      ('View management screens',('manage_tabs','manage_main')),
      ('Edit Properties',('manage_editForm','manage_edit')),
      ('Change permissions',('manage_access',)),
      ('View',('__call__','')),
  )

  def __init__(self,id,title,DocToParse,TTL,CacheSlots):
    self.id = id
    self.title = title
    self.TTL = TTL
    self.DocToParse = DocToParse
    self.CacheSlots = CacheSlots
    self.Cache = []
    self._v_lock = allocate_lock()

  def __setstate__(self,state):
    Persistent.__setstate__(self,state)
    self._v_lock = allocate_lock()

  def __call__(self, client=None, REQUEST={}, RESPONSE=None, **kw):
    """Process cache slots"""
    kw['document_id']   =self.id
    kw['document_title']=self.title

    # Verify the authenticated user object.
    try:
      if REQUEST.has_key('AUTHENTICATED_USER'):
        verify_watermark(REQUEST['AUTHENTICATED_USER'])
    except:
      #Where on zope 2.2, works anyway
      pass

    #Is this folder cacheable?
    if hasattr(self,'toCache') and getattr(self,'toCache'):
      #Find if document exists in cache
      pos = self.findInCache()
    else:
      pos = -1


    #If there's no more place...
    if pos == -1:

      r = apply(HTML.__call__, (eval('self.aq_parent.' + self.DocToParse), client, REQUEST), kw)
      return r

#    self._v_lock.acquire()
#    try:
    self.Cache[pos].hpcc = self.Cache[pos].hpcc  + 1
#    finally:
#      self._v_lock.release()

    if ((int(self.Cache[pos].CacheDate) + int(self.TTL)) <= time.time()):
      """Invalidate CacheSlot"""
      self.Cache[pos].Cached = None

    if self.Cache[pos].Cached == None:
      """Refresh CacheSlot"""
#      self._v_lock.acquire()
#      try:
#      self._p_jar.sync()
#        get_transaction().commit(1)
      self.Cache[pos].Last_hpcc = self.Cache[pos].hpcc
      self.Cache[pos].hpcc = 1
      self.Cache[pos].QueryString = self.REQUEST['QUERY_STRING']
      self.Cache[pos].URL1 = self.REQUEST['URL1']
      self.Cache[pos].CacheData = apply(HTML.__call__, (eval('self.aq_parent.' + self.DocToParse), client, REQUEST), kw)
      self.Cache[pos].CacheDate = int(time.time())
      self.Cache[pos].Cached = 1
#      finally:
#      self._p_changed = 1
#      self._v_lock.release()

    return self.Cache[pos].CacheData

  def findInCache(self):
    """Search the cacheslot pool for an instance with the same URL1 and QueryString.
        If none is found and there are free slots, add a fresh one, else return -1"""
    for pos in range(len(self.Cache)):
      # Try to find one with the same properties
      if (self.Cache[pos].URL1 == self.REQUEST['URL1']) and (self.Cache[pos].QueryString == self.REQUEST['QUERY_STRING']):
	return pos
    if len(self.Cache) < int(self.CacheSlots):
      # Add a slot if space is available
      self.Cache.append(CacheSlot())
      return len(self.Cache) - 1
    # Find expired with the lowest hppc, bigger than 0(coz these are new, uncycled)
    Expired = []
    for pos in range(len(self.Cache)):
      if (self.Cache[pos].hpcc < 1) or ((int(self.Cache[pos].CacheDate) + int(self.TTL)) > time.time()):
        Expired.append(9999999)
      else:
        Expired.append(self.Cache[pos].hpcc)
    if min(Expired) < 9999999:
      del self.Cache[(Expired.index(min(Expired)))]
      self.Cache.append(CacheSlot())
      return len(self.Cache) - 1
    return -1


  def now(self):
    """Now"""
    return time.time()

  def flush_cache(self,Slot=None,REQUEST=None):
    """Flush some cache slots"""
    if Slot == None:
      self.Cache = []
    else:
      try:
        Slot = split(Slot)
      except:
        pass
      Slot.reverse()
      for item in Slot:
        del self.Cache[int(item)]
    if REQUEST is not None:
        return MessageDialog(
            title='Flush Cache',
            message='<strong>%s</strong>\'s caches have been flushed.' %
            self.id,
            action ='./manage_viewCache',
              )

  manage_viewCache = HTMLFile('CachePoolView',globals())

  ######################################################################
  #EDIT STUFF
  manage_editForm = HTMLFile('CachePoolEdit',globals())
  manage_main = manage_editForm
  def manage_edit(self,title,TTL,DocToParse,CacheSlots,REQUEST=None):
      "edit Cache Pool characteristics"
      self.title=title
      self.TTL = TTL
      self.DocToParse = DocToParse
      if CacheSlots < self.CacheSlots:
        self.Cache = self.Cache[:int(CacheSlots)]
      self.CacheSlots = CacheSlots
      if REQUEST is not None:
          return MessageDialog(
              title='Edited',
              message='<strong>%s</strong> has been edited.' %
              self.id,
              action ='./manage_editForm',
                )



class CacheSlot:
  """Cache Slot Definition"""

  def __init__(self):
    self.URL1 = ''
    self.QueryString = ''
    self.hpcc = 0
    self.Last_hppc = -1
    self.CacheData = ''
    self.CacheDate = 0
    self.Cached = None
